/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:  
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "WSIF" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 2001, 2002, International
 * Business Machines, Inc., http://www.apache.org.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

package org.apache.wsif.naming;

import javax.naming.NamingException;
import javax.naming.Reference;
import javax.naming.Referenceable;
import javax.naming.StringRefAddr;
import org.apache.wsif.logging.Trc;

/**
 * This is a lightweight object which provides a reference for a service stub. 
 * When passed to Context.bind(), the getReference() method is invoked and the 
 * resulting Reference object is stored in the directory by JNDI.
 *  
 * @author Owen Burroughs <owenb@apache.org>
 */
public class WSIFServiceStubRef implements Referenceable {

    String wsdlLoc;
    String serviceNS;
    String serviceName;
    String portTypeNS;
    String portTypeName;
    String preferredPort;
    String className;

    /**
     * Constructor that takes all necessary information needed to create a 
     * WSIFService and stub.
     * @param wsdl The location of the wsdl file
     * @param sNS The namespace for the service as specified in the wsdl
     * @param sName The name of the service required, as specified in the wsdl
     * @param ptNS The namespace of the port type required, as specified in the wsdl
     * @param ptName The name of the port type required, as specified in the wsdl
     * @param portName The name of the preferred port to use
     * @param cls The fully qualified name of the interface class for the stub
     */
    public WSIFServiceStubRef(
            String wsdl,
            String sNS,
            String sName,
            String ptNS,
            String ptName,
            String portName,
            String cls) {
        Trc.entry(this, wsdl, sNS, sName, ptNS, ptName, portName, cls);

        wsdlLoc = wsdl;
        serviceNS = sNS;
        serviceName = sName;
        portTypeNS = ptNS;
        portTypeName = ptName;
        preferredPort = portName;
        className = cls;
        if (Trc.ON)
            Trc.exit(deep());
    }

    /**
     * Method to create and return a Reference object for the service.
     * @return A Reference object containing the information required to create a 
     * WSIFService and return it, when a lookup is performed on the service using
     * JNDI.
     */
    public Reference getReference() throws NamingException {
        Trc.entry(this);

        Reference ref =
            new Reference(
                WSIFServiceStubRef.class.getName(),
                WSIFServiceObjectFactory.class.getName(),
                null);
        ref.add(new StringRefAddr("wsdlLoc", wsdlLoc));
        ref.add(new StringRefAddr("serviceNS", serviceNS));
        ref.add(new StringRefAddr("serviceName", serviceName));
        ref.add(new StringRefAddr("portTypeNS", portTypeNS));
        ref.add(new StringRefAddr("portTypeName", portTypeName));
        ref.add(new StringRefAddr("preferredPort", preferredPort));
        ref.add(new StringRefAddr("className", className));

        Trc.exit(ref);
        return ref;
    }

    public String deep() {
        String buff = "";
        try {
            buff = new String(this.toString() + "\n");
            buff += "wsdlLoc: " + wsdlLoc;
            buff += " serviceNS: " + serviceNS;
            buff += " serviceName: " + serviceName;
            buff += " portTypeNS: " + portTypeNS;
            buff += " portTypeName: " + portTypeName;
            buff += " preferredPort: " + preferredPort;
            buff += " className: " + className;
        } catch (Exception e) {
            Trc.exceptionInTrace(e);
        }
        return buff;
    }
}