/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:  
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "WSIF" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 2001, 2002, International
 * Business Machines, Inc., http://www.apache.org.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

package org.apache.wsif.providers.java;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import javax.wsdl.BindingFault;
import javax.wsdl.BindingOperation;
import javax.wsdl.BindingOutput;
import javax.wsdl.Fault;
import javax.wsdl.Operation;
import javax.wsdl.OperationType;
import javax.wsdl.Part;
import javax.wsdl.Port;
import javax.xml.namespace.QName;

import org.apache.wsif.WSIFException;
import org.apache.wsif.WSIFMessage;
import org.apache.wsif.WSIFOperation;
import org.apache.wsif.WSIFPort;
import org.apache.wsif.base.WSIFDefaultOperation;
import org.apache.wsif.logging.MessageLogger;
import org.apache.wsif.logging.Trc;
import org.apache.wsif.providers.ProviderUtils;
import org.apache.wsif.wsdl.extensions.java.JavaOperation;

/**
 * Java operation. 
 * @see WSIFPort_Java
 * @author <a href="mailto:gpfau@de.ibm.com">Gerhard Pfau</a>
 * @author Owen Burroughs <owenb@apache.org>
 * @author Jeremy Hughes <hughesj@apache.org>
 */
public class WSIFOperation_Java
        extends WSIFDefaultOperation
        implements WSIFOperation {
        	
    private static final long serialVersionUID = 1L;
        	
    protected javax.wsdl.Port fieldPortModel;
    protected WSIFPort_Java fieldPort;
    protected javax.wsdl.BindingOperation fieldBindingOperationModel;
    protected JavaOperation fieldJavaOperationModel;
    
    protected String[] fieldInParameterNames = null;
    protected String[] fieldOutParameterNames = null;
    protected Map fieldFaultMessageInfos = null;
    protected Method[] fieldMethods = null;
    protected Constructor[] fieldConstructors = null;
    protected String fieldOutputMessageName = null;
    protected boolean fieldIsStatic = false;
    protected boolean fieldIsConstructor = false;
    protected Map fieldTypeMaps = null;

    private class FaultMessageInfo {
        String fieldMessageName;
        String fieldPartName;
        String fieldFormatType;
        // Note: In Java fault messages contain only one part: the Java exception

        FaultMessageInfo(String messageName, String partName, String formatType) {
            fieldMessageName = messageName;
            fieldPartName = partName;
            fieldFormatType = formatType;
        }
    }

    public WSIFOperation_Java(
        javax.wsdl.Port portModel,
        BindingOperation bindingOperationModel,
        WSIFPort_Java port,
        Map typeMaps)
        throws WSIFException {

        Trc.entry(this, portModel, bindingOperationModel, port, typeMaps);

        fieldPortModel = portModel;
        fieldBindingOperationModel = bindingOperationModel;
        fieldPort = port;
        fieldTypeMaps = typeMaps;
        Method[] allMethods = fieldPort.getObjectReference().getClass().getMethods();

        try {
            fieldJavaOperationModel =
                (JavaOperation) fieldBindingOperationModel.getExtensibilityElements().get(0);
        } catch (Exception e) {
        	Trc.exception(e);
            throw new WSIFException(
                "Unable to resolve Java binding for operation '"
                    + bindingOperationModel.getName()
                    + "'");
        }

        // Check what kind of method we have...
        String methodType = fieldJavaOperationModel.getMethodType();
        if (methodType != null) {
            if (methodType.equals("static")) {
                fieldIsStatic = true;
            } else if (methodType.equals("constructor")) {
                fieldIsConstructor = true;
            } else {
                // Assume instance method...
            }
        }

        fieldConstructors = getConstructors();
        fieldMethods = getMethods(allMethods);

        if (Trc.ON)
            Trc.exit(deep());
    }

    private WSIFOperation_Java(
        Port p,
        WSIFPort_Java pj,
        BindingOperation bo,
        JavaOperation jo,
        String[] inPNames,
        String[] outPNames,
        Map faultMsgInfos,
        Method[] m,
        Constructor[] c,
        String outMName,
        boolean isSttc,
        boolean isCnstr,
        Map tMap) {
        Trc.entry(
            this,
            p,
            pj,
            bo,
            jo,
            inPNames,
            outPNames,
            faultMsgInfos,
            m,
            c,
            outMName,
            new Boolean(isSttc),
            new Boolean(isCnstr),
            tMap);

        fieldPortModel = p;
        fieldPort = pj;
        fieldBindingOperationModel = bo;
        fieldJavaOperationModel = jo;
        fieldInParameterNames = inPNames;
        fieldOutParameterNames = outPNames;
        fieldFaultMessageInfos = faultMsgInfos;
        fieldMethods = m;
        fieldConstructors = c;
        fieldOutputMessageName = outMName;
        fieldIsStatic = isSttc;
        fieldIsConstructor = isCnstr;
        fieldTypeMaps = tMap;

        if (Trc.ON)
            Trc.exit(deep());
    }
    
    /**
     * Create a new copy of this object. This is not a clone, since
     * it does not copy the referenced objects as well.
     */
    public WSIFOperation_Java copy() throws WSIFException {
        Trc.entry(this);
        WSIFOperation_Java woj =
            new WSIFOperation_Java(
                fieldPortModel,
                fieldPort,
                fieldBindingOperationModel,
                fieldJavaOperationModel,
                fieldInParameterNames,
                fieldOutParameterNames,
                fieldFaultMessageInfos,
                fieldMethods,
                fieldConstructors,
                fieldOutputMessageName,
                fieldIsStatic,
                fieldIsConstructor,
                fieldTypeMaps);
        Trc.exit(woj);
        return woj;
    }

    protected static Class getClassForName(String classname) throws WSIFException {
    	Trc.entry(null,classname);
    	
        Class cls = null;

        if (classname == null) {
            throw new WSIFException("Error in getClassForName(): No class name specified!");
        }

        try {
            if (classname.lastIndexOf('.') == -1) {
                // Have to check for built in data types
                if (classname.equals("char")) {
                    cls = char.class;
                } else if (classname.equals("boolean")) {
                    cls = boolean.class;
                } else if (classname.equals("byte")) {
                    cls = byte.class;
                } else if (classname.equals("short")) {
                    cls = short.class;
                } else if (classname.equals("int")) {
                    cls = int.class;
                } else if (classname.equals("long")) {
                    cls = long.class;
                } else if (classname.equals("float")) {
                    cls = float.class;
                } else if (classname.equals("double")) {
                    cls = double.class;
                } else {
                    // Load the class using the Thread context's class loader
                    cls =
                        Class.forName(classname, true, Thread.currentThread().getContextClassLoader());
                }
            } else {
                cls =
                    Class.forName(classname, true, Thread.currentThread().getContextClassLoader());
            }
        } catch (ClassNotFoundException ex) {
        	Trc.exception(ex);
            throw new WSIFException("Could not instantiate class '" + classname + "'", ex);
        }
        Trc.exit(cls);
        return cls;
    }

    protected Constructor[] getConstructors() throws WSIFException {
    	Trc.entry(this);
        Constructor[] candidates;
        // Get the possible constructors with the argument classes we've found.
        Constructor[] constructors =
            fieldPort.getObjectReference().getClass().getConstructors();
        Object[] args = getMethodArgumentClasses();
        Vector possibles = new Vector();
        for (int i = 0; i < constructors.length; i++) {
            Class[] params = constructors[i].getParameterTypes();
            if (params.length != args.length)
                continue;

            boolean match = true;
            for (int j = 0; j < params.length; j++) {
                Object obj = args[j];
                if (obj instanceof Vector) {
                    Vector vec = (Vector) obj;
                    boolean found = false;
                    for (int p = 0; p < vec.size(); p++) {
                        Class cl = (Class) vec.get(p);
                        if (cl.getName().equals(params[j].getName())) {
                            found = true;
                            break;
                        } else if (params[j].isAssignableFrom(cl)) {
                            found = true;
                            break;
                        }
                    }
                    if (!found) {
                        match = false;
                        break;
                    }
                } else {
                    if (!(((Class) obj).getName().equals(params[j].getName()))
                        && !(params[j].isAssignableFrom((Class) obj))) {
                        match = false;
                        break;
                    }
                }
            }
            if (match) {
                possibles.addElement(constructors[i]);
            }
        }
        candidates = new Constructor[possibles.size()];
        for (int k = 0; k < candidates.length; k++) {
            candidates[k] = (Constructor) possibles.get(k);
        }
        Trc.exit(candidates);
        return candidates;
    }

    protected Map getFaultMessageInfos() throws WSIFException {
    	Trc.entry(this);
        // Get the current operation
        Operation operation = null;
        try {
            operation = getOperation();
        } catch (Exception e) {
        	Trc.exception(e);
            throw new WSIFException("Failed to get Operation", e);
        }

        if (fieldFaultMessageInfos == null) {
            fieldFaultMessageInfos = new HashMap();
        }

        BindingFault bindingFaultModel = null;
        Map bindingFaultModels = fieldBindingOperationModel.getBindingFaults();
        List parts = null;
        Iterator modelsIterator = bindingFaultModels.values().iterator();

        while (modelsIterator.hasNext()) {
            bindingFaultModel = (BindingFault) modelsIterator.next();
            String name = bindingFaultModel.getName();
            if (name == null) {
                throw new WSIFException("Fault name not found in binding");
            }

            Map map = operation.getFault(name).getMessage().getParts();
            if (map.size() >= 1) {
                Part part = (Part) map.values().iterator().next();
                QName partType = part.getTypeName();
                if (partType == null) partType = part.getElementName();
                Object formatType = fieldTypeMaps.get(partType);
                if (formatType == null) {
                    throw new WSIFException(
                        "formatType for typeName '" + part.getName() + "' not found in document");
                }

                if (formatType instanceof Vector) {
                    Vector types = (Vector) formatType;
                    Enumeration enum = types.elements();
                    while (enum.hasMoreElements()) {
                        String type = (String) enum.nextElement();
                        // Add new fault message information to the map
                        fieldFaultMessageInfos.put(
                            type,
                            new FaultMessageInfo(name, part.getName(), type));
                    }
                } else {
                    String type = (String) formatType;
                    // Add new fault message information to the map
                    fieldFaultMessageInfos.put(
                        type,
                        new FaultMessageInfo(name, part.getName(), type));
                }
            }
        }
        Trc.exit(fieldFaultMessageInfos);
        return fieldFaultMessageInfos;
    }

    protected Method[] getMethods(Method[] allMethods) throws WSIFException {
    	Trc.entry(this,allMethods);
        Method[] candidates;
        try {
            if (!fieldIsConstructor) {
                // Get the possible methods with the argument classes we've found.
                Object[] args = getMethodArgumentClasses();
                Object retClass = getMethodReturnClass();
                Vector possibles = new Vector();
                for (int i = 0; i < allMethods.length; i++) {
                    if (!allMethods[i]
                        .getName()
                        .equals(fieldJavaOperationModel.getMethodName()))
                        continue;
                    Class[] params = allMethods[i].getParameterTypes();
                    if (params.length != args.length)
                        continue;
                    Class retType = allMethods[i].getReturnType();
                    if (retClass != null && retClass instanceof Vector) {
                        Vector vec = (Vector) retClass;
                        boolean found = false;
                        for (int p = 0; p < vec.size(); p++) {
                            Class cl = (Class) vec.get(p);
                            if (cl.getName().equals(retType.getName())) {
                                found = true;
                                break;
                            } else if (cl.isAssignableFrom(retType)) {
                              	found = true;
                               	break;
                            }
                        }
                        if (!found)
                            continue;
                    } else {
                        if (retType != null && retClass != null) {
                            if (!(((Class) retClass).getName().equals(retType.getName()))
                            	&& !(((Class) retClass).isAssignableFrom(retType)))
                                continue;
                        }
                    }

                    boolean match = true;
                    for (int j = 0; j < params.length; j++) {
                        Object obj = args[j];
                        if (obj instanceof Vector) {
                            Vector vec = (Vector) obj;
                            boolean found = false;
                            for (int p = 0; p < vec.size(); p++) {
                                Class cl = (Class) vec.get(p);
                                if (cl.getName().equals(params[j].getName())) {
                                    found = true;
                                    break;
                                } else if (params[j].isAssignableFrom(cl)) {
                                	found = true;
                                	break;
                                }
                            }
                            if (!found) {
                                match = false;
                                break;
                            }
                        } else {
                            if (!(((Class) obj).getName().equals(params[j].getName()))
                            	&& !(params[j].isAssignableFrom((Class) obj))) {
                                match = false;
                                break;
                            }
                        }
                    }
                    if (match) {
                        possibles.addElement(allMethods[i]);
                    }
                }
                candidates = new Method[possibles.size()];
                for (int k = 0; k < candidates.length; k++) {
                    candidates[k] = (Method) possibles.get(k);
                }
                Trc.exit(candidates);
                return candidates;
            }
            Method[] m = new Method[0];
            Trc.exit(m);
            return m;
        } catch (WSIFException ex) {
        	Trc.exception(ex);
            throw ex;
        } catch (Exception ex) {
        	Trc.exception(ex);
            throw new WSIFException(
                "Error while resolving meta information of method "
                    + fieldJavaOperationModel.getMethodName()
                    + " : The meta information is not consistent.",
                ex);
        }
    }

    protected Operation getOperation() throws Exception {
    	Trc.entry(this);
        Operation operation = null;

        // <input> and <output> tags in binding operations are not mandatory
        // so deal with null BindingInputs or BindingOutputs
        String inputName = null;
        if (fieldBindingOperationModel.getBindingInput()!=null)
            inputName = fieldBindingOperationModel.getBindingInput().getName();

        String outputName = null;
        if (fieldBindingOperationModel.getBindingOutput()!=null)
            outputName = fieldBindingOperationModel.getBindingOutput().getName();

        // Build the parts list
        //this.fieldBindingOperationModel.getBindingInput().getName(),
        operation =
            this.fieldPortModel.getBinding().getPortType().getOperation(
                this.fieldBindingOperationModel.getName(),
                inputName,
                outputName);
        Trc.exit(operation);
        return operation;
    }

    protected Object getMethodReturnClass() throws WSIFException {
    	Trc.entry(this);
        Object methodReturnClass = null;
        try {
            String returnPartString = fieldJavaOperationModel.getReturnPart();
            if (returnPartString != null) {
                // A returnPart has been specified so check that this method has the correct
                // return type
                Part returnPart =
                    getOperation().getOutput().getMessage().getPart(returnPartString);

                // If there is no returnPart specified then not interested in return value
                if (returnPart != null) {
                    QName partType = returnPart.getTypeName();                    
                    if (partType == null) partType = returnPart.getElementName();
                    Object obj = this.fieldTypeMaps.get(partType);
                    if (obj == null)
                        throw new WSIFException(
                            "Could not map type "
                                + partType
                                + " to a java type. Part name was "
                                + returnPart.getName() == null 
                                    ? "<null>" : returnPart.getName());
                                
                    if (obj instanceof Vector) {
                        Vector v = (Vector) obj;
                        Vector argv = new Vector();
                        Enumeration enum = v.elements();
                        while (enum.hasMoreElements()) {
                            String cls = (String) enum.nextElement();
                            argv.addElement(getClassForName(cls));
                        }
                        methodReturnClass = argv;
                    } else {
                        methodReturnClass =
                            getClassForName((String) obj);
                    }
                } else {
                    // If we get here then the return part specified on the java operation was not
                    // in the output message
                    throw new Exception(
                        "returnPart '" + returnPartString + "' was not in the output message");
                }
            }
            // returnPart attribute was not present so return methodReturnClass as default null
        } catch (Exception ex) {
        	Trc.exception(ex);
            throw new WSIFException(
                "Error while determining return class of method "
                    + fieldJavaOperationModel.getMethodName()
                    + " : The meta information is not consistent.",
                ex);
        }
        Trc.exit(methodReturnClass);
        return methodReturnClass;
    }

    protected Object[] getMethodArgumentClasses() throws WSIFException {
    	Trc.entry(this);
        Object[] methodArgClasses = null;
        try {

            Operation operation = getOperation();

            /*
            The order of the parameters as passed to the Java method is
            determined in this way:
            1. from the <wsdl:operation parameterOrder="xxx yyy"> attribute if present
            2. from the <java:operation parameterOrder="xxx yyy"> attribute if present
            3. from the order the parts are specified in the input message
            */

            // Get the parameter order according to the above rules
            List parameterOrder = null;

            parameterOrder = fieldJavaOperationModel.getParameterOrder();

            if (parameterOrder == null) {
                parameterOrder = operation.getParameterOrdering();
            }

            /*
            The MessageImpl actually the order the parts were added in a List, but
            this can only be accessed via the getOrderedParts() methods which
            of course returns a List of Parts!
            So here I (rather inefficiently) convert the list of parts to a list of
            string part names.
            */
            if (parameterOrder == null) {
                List partList = operation.getInput().getMessage().getOrderedParts(null);
                parameterOrder = new Vector();
                Iterator partListIterator = partList.iterator();
                while (partListIterator.hasNext()) {
                    Part part = (Part) partListIterator.next();
                    parameterOrder.add(part.getName());
                }
            }

            /*
            	Operations do not specify whether they are to be used with RPC-like bindings 
            	or not. However, when using an operation with an RPC-binding, it is useful to 
            	be able to capture the original RPC function signature. For this reason, 
            	a request-response or solicit-response operation MAY specify a list of parameter
            	names via the parameterOrder attribute (of type nmtokens). The value of the 
            	attribute is a list of message part names separated by a single space. 
            	The value of the parameterOrder attribute MUST follow the following rules:
            
                * The part name order reflects the order of the parameters in the RPC signature
                * The return value part is not present in the list
                * If a part name appears in both the input and output message, it is an in/out parameter
            	* If a part name appears in only the input message, it is an in parameter
            	* If a part name appears in only the output message, it is an out parameter
            
            	Note that this information serves as a "hint" and may safely be ignored by 
            	those not concerned with RPC signatures. Also, it is not required to be present, 
            	even if the operation is to be used with an RPC-like binding.
            */

            ArrayList argNames = new ArrayList();
            ArrayList argTypes = new ArrayList();

            Iterator parameterIterator = parameterOrder.iterator();
            while (parameterIterator.hasNext()) {
                String param = (String) parameterIterator.next();
                Part part = (Part) operation.getInput().getMessage().getPart(param);
                if (part == null) {
                    part = (Part) operation.getOutput().getMessage().getPart(param);
                }
                if (part == null)
                    throw new Exception(
                        "Part '"
                            + param
                            + "' from parameterOrder not found in input or output message");
                argNames.add((String) part.getName());

                // should also check for the element
                QName partType = part.getTypeName();;
                if (partType == null) partType = part.getElementName();                
                Object obj = this.fieldTypeMaps.get(partType);
                if (obj == null)
                    throw new WSIFException(
                        "Could not map type "
                            + partType
                            + " to a java type. Part name was "
                            + part.getName() == null ? "<null>" : part.getName());
                
                if (obj instanceof Vector) {
                    Vector v = (Vector) obj;
                    Vector argv = new Vector();
                    Enumeration enum = v.elements();
                    while (enum.hasMoreElements()) {
                        String cls = (String) enum.nextElement();
                        argv.addElement(getClassForName(cls));
                    }
                    argTypes.add(argv);
                } else {
                    argTypes.add(getClassForName((String)obj));
                }

            }

            methodArgClasses = argTypes.toArray();
            fieldInParameterNames = new String[argNames.size()];
            for (int i = 0; i < argNames.size(); i++) {
                fieldInParameterNames[i] = (String) argNames.get(i);
            }

            // Deal with output parts if operation is Request-Response
            if (operation.getStyle().equals(OperationType.REQUEST_RESPONSE)) {
                argNames = new ArrayList();
                // Get the returnPart attribute if it exists
                String returnPart = fieldJavaOperationModel.getReturnPart();                
                Iterator outputPartsIterator =
                    operation.getOutput().getMessage().getOrderedParts(null).iterator();
                while (outputPartsIterator.hasNext()) {
                    Part part = (Part) outputPartsIterator.next();
                    String partName = part.getName();
                    if (partName != null && returnPart != null && partName.equals(returnPart)) {
                    	// Put return part first in the list of output parts
                    	argNames.add(0, partName);
                    } else {                    
                        argNames.add((String) part.getName());
                    }
                }                               

                // Populate an array of output message part names
				fieldOutParameterNames = new String[argNames.size()];
                for (int i = 0; i < argNames.size(); i++) {
                    fieldOutParameterNames[i] = (String) argNames.get(i);
                }
            } else {
            	fieldOutParameterNames = new String[0];
            }
        } catch (Exception ex) {
        	Trc.exception(ex);
            throw new WSIFException(
                "Error while determining signature of method "
                    + fieldJavaOperationModel.getMethodName()
                    + " : The meta information is not consistent.",
                ex);
        }
        Trc.exit(methodArgClasses);
        return methodArgClasses;
    }

    // Turns an array of arguments into a form compatible with a method
    // If they are compatible, the object array is populated
    // otherwise returns null
    protected Object[] getCompatibleArguments(Class[] parmTypes, Object[] args) {
    	Trc.entry(this,parmTypes,args);
        // Go through each argument checking it's compatability with the method arg
        // creating a compatible set along the way.
        // In essence this just converts from String to Character when necessary
        // if there are further special case classes such as these which are dependent
        // on the object value, PUT THEM HERE :-)
        if (args == null || parmTypes == null) {
            Object[] compatibleArgs = new Object[0];
            return compatibleArgs;
        }

        Object[] compatibleArgs = new Object[args.length];
        for (int i = 0; i < parmTypes.length; i++) {
            // If the arg is a null then skip it
            if (args[i] == null) {
                compatibleArgs[i] = ProviderUtils.getDefaultObject(parmTypes[i]);
                continue;
            }
            // Consider the special cas, squeezing a String into a Character
            Object convertedArg = getCompatibleObject(parmTypes[i], args[i]);
            if (convertedArg == null) {
                // can't convert one of the arguments so return null
                return null;
            } else {
                compatibleArgs[i] = convertedArg;
            }

        }
        Trc.exit(compatibleArgs);
        return compatibleArgs;
    }

    protected Object getCompatibleReturn(Method method, Object returnObj) {
    	Trc.entry(this,method,returnObj);
    	Object o = null;
    	Class rt = method.getReturnType();
    	Class ct = null;
    	int dims = 0;
    	if (rt.isArray()) {
    		ct = rt.getComponentType();
    		dims++;
        	while (ct.isArray()) {
        		ct = ct.getComponentType();
        		dims++;
        	}
    	}
        if (returnObj instanceof java.lang.Character) {
            o = getCompatibleObject(java.lang.String.class, returnObj);
        } else if (ct != null && (ct.equals(java.lang.Character.class) || ct.equals(char.class))) {
        	String stringArrayClassName = "[Ljava.lang.String;";
        	for (int d=1; d<dims; d++) {
        		stringArrayClassName = "["+stringArrayClassName;
        	}
        	try {
        		Class stringArrayClass = Class.forName(stringArrayClassName, true, Thread.currentThread().getContextClassLoader());
        		o = getCompatibleObject(stringArrayClass, returnObj);
        	} catch(ClassNotFoundException cnf) {
        		Trc.ignoredException(cnf);
        		o = returnObj;
        	}
        } else {
            o = returnObj;
        }
        Trc.exit(o);
        return o;
    }

    // Usually cls1.isAssignableFrom(cls2) returning false means you can't cast 
    // instance of cls1 to cls2. There are some special cases we need to cover ...
    // String->Character and Character->String
    // If a conversion is known then the obj is converted to class cls
    //   if that conversion failed, null is returned
    //   else the converted obj is returned
    // If a conversion is not known about then the obj is returned
    // Note: if you are adding other cases ensure you add both directions since the
    //       this conversion may be needed on method args AND returns
    protected Object getCompatibleObject(Class cls, Object obj) {
    	Trc.entry(this,cls,obj);
    	
    	if (cls.getName().equals(obj.getClass().getName())) return obj;
    	  	
        // String -> Character
        if ((cls.equals(java.lang.Character.class) || cls.equals(char.class))
            && obj.getClass().equals(java.lang.String.class)) {
            Character charArg = ProviderUtils.stringToCharacter((String) obj);
            if (charArg == null) {
                // Can't convert this string to character so return null
                Trc.exit(null);
                return null;
            }
            Trc.exit(charArg);
            return charArg;
        }
        
        // String arrays -> char/Character arrays and Character arrays -> String arrays
        if (cls.isArray() && obj.getClass().isArray()) {
        	Class cct = cls.getComponentType();
        	Class objct = obj.getClass().getComponentType();
        	while (cct.isArray()) {
        		cct = cct.getComponentType();
        	}
        	while (objct.isArray()) {
        		objct = objct.getComponentType();
        	}         	
        	if (objct.equals(java.lang.String.class) && cct.equals(char.class)) {
        		try {
        			Object charArray = ProviderUtils.stringArrayToCharArray(obj);
        			Trc.exit(charArray);
        			return charArray;
        		} catch (Exception e) {
        			Trc.ignoredException(e);
            		Trc.exit(null);
            		return null;
        		}
        	} else if (objct.equals(java.lang.String.class) && cct.equals(java.lang.Character.class)) {
        		try {
        			Object charArray = ProviderUtils.stringArrayToCharacterArray(obj);
        			Trc.exit(charArray);
        			return charArray;
        		} catch (Exception e) {
        			Trc.ignoredException(e);
            		Trc.exit(null);
            		return null;
        		}
        	} else if (objct.equals(java.lang.Character.class) && cct.equals(java.lang.String.class)) {
        		try {
        			Object charArray = ProviderUtils.characterArrayToStringArray(obj);
        			Trc.exit(charArray);
        			return charArray;
        		} catch (Exception e) {
        			Trc.ignoredException(e);
            		Trc.exit(null);
            		return null;
        		}
        	} else if (objct.equals(char.class) && cct.equals(java.lang.String.class)) {
        		try {
        			Object charArray = ProviderUtils.charArrayToStringArray(obj);
        			Trc.exit(charArray);
        			return charArray;
        		} catch (Exception e) {
        			Trc.ignoredException(e);
            		Trc.exit(null);
            		return null;
        		}
        	}
        }
        
        if (cls.equals(java.lang.String.class)
            && obj.getClass().equals(java.lang.Character.class)) {
            Trc.exit(obj.toString());
            return (obj.toString());
        }
        Trc.exit(obj);
        return obj;
    }

    protected String getOutputMessageName() throws WSIFException {
    	Trc.entry(this);
        if (fieldOutputMessageName == null) {
            BindingOutput bindingOutputModel =
                fieldBindingOperationModel.getBindingOutput();
            if (bindingOutputModel != null) {
                fieldOutputMessageName = bindingOutputModel.getName();
            }
        }
        Trc.exit(fieldOutputMessageName);
        return fieldOutputMessageName;
    }

    public WSIFPort getWSIFPort() {
        Trc.entry(this);
        Trc.exit(fieldPort);
        return fieldPort;
    }
    
    public boolean executeRequestResponseOperation(
        WSIFMessage input,
        WSIFMessage output,
        WSIFMessage fault)
        throws WSIFException {

        Trc.entry(this, input, output, fault);
        close();

        boolean operationSucceeded = true;
        boolean usedOutputParam = false;
        Method chosenMethod = null;

        try {
            Object result = null;
            // Need to get the stuff here because this also initializes fieldInParameterNames
            if (fieldIsConstructor) {
                if (fieldConstructors.length <= 0)
                    throw new WSIFException(
                        "No constructor found that match the parts specified");
            } else {
                if (fieldMethods.length <= 0)
                    throw new WSIFException("No method named '"
                        + fieldJavaOperationModel.getMethodName()
                        + "' found that match the parts specified");
            }

            Object[] arguments = null;
            Object part = null;
            if ((fieldInParameterNames != null) && (fieldInParameterNames.length > 0)) {
                arguments = new Object[fieldInParameterNames.length];
                for (int i = 0; i < fieldInParameterNames.length; i++) {
                    try {
                        part = input.getObjectPart(fieldInParameterNames[i]);
                        arguments[i] = part;
                    } catch (WSIFException e) {
                     	Trc.exception(e);
                        arguments[i] = null;
                        if (fieldOutParameterNames.length > 0) {
                            String outParameterName = null;
                            for (int j = 0; j < fieldOutParameterNames.length; j++) {
                                outParameterName =
                                    fieldOutParameterNames[j];
                                if ((outParameterName != null)
                                    && (outParameterName
                                        .equals(fieldInParameterNames[i]))) {
                                    arguments[i] =
                                        (fieldMethods[0].getParameterTypes()[i])
                                            .newInstance();
                                    usedOutputParam = true;
                                }
                            }
                        }
                    }
                }
            }

            boolean invokedOK = false;
            if (fieldIsConstructor) {
                for (int a = 0; a < fieldConstructors.length; a++) {
                    try {
                        // Get a set of arguments which are compatible with the ctor
                        Object[] compatibleArguments =
                            getCompatibleArguments(fieldConstructors[a].getParameterTypes(), arguments);
                        // If we didn't get any arguments then the parts aren't compatible with the ctor
                        if (compatibleArguments == null)
                            break;
                        // Parts are compatible so invoke the ctor with the compatible set

                        Trc.event(
                            this,
                            "Invoking constructor ",
                            fieldConstructors[a],
                            " with arguments ",
                            compatibleArguments);

                        result =
                            fieldConstructors[a].newInstance(
                                compatibleArguments);

                        Trc.event(
                            this,
                            "Returned from constructor, result is ",
                            result);
                        
                        fieldPort.setObjectReference(result);
                        invokedOK = true;
                        break;
                    } catch (IllegalArgumentException ia) {
                    	Trc.ignoredException(ia);
                        // Ignore and try next constructor
                    }
                }
                if (!invokedOK)
                    throw new WSIFException("Failed to call constructor for object in Java operation");
                // Side effect: Initialize port's object reference
            } else {
                if (fieldIsStatic) {
                    for (int a = 0; a < fieldMethods.length; a++) {
                        if (usedOutputParam) {
                            for (int i = 0; i < fieldInParameterNames.length; i++) {
                                String outParameterName = null;
                                for (int j = 0; j < fieldOutParameterNames.length; j++) {
                                    outParameterName = fieldOutParameterNames[j];
                                    if ((outParameterName != null)
                                        && (outParameterName.equals(fieldInParameterNames[i]))) {
                                        arguments[i] = (fieldMethods[a].getParameterTypes()[i]).newInstance();
                                    }
                                }
                            }
                        }
                        try {
                            // Get a set of arguments which are compatible with the method
                            Object[] compatibleArguments =
                                getCompatibleArguments(fieldMethods[a].getParameterTypes(), arguments);
                            // If we didn't get any arguments then the parts aren't compatible with the method
                            if (compatibleArguments == null)
                                break;
                            // Parts are compatible so invoke the method with the compatible set

                            Trc.event(
                                this,
                                "Invoking method ",
                                fieldMethods[a],
                                " with arguments ",
                                compatibleArguments);

                            result =
                                fieldMethods[a].invoke(
                                    null,
                                    compatibleArguments);

                            Trc.event(
                                this,
                                "Returned from method, result is ",
                                result);

                            chosenMethod = fieldMethods[a];
                            invokedOK = true;
                            break;
                        } catch (IllegalArgumentException ia) {
                           	Trc.ignoredException(ia);
                            // Ignore and try next method
                        }
                    }
                    if (!invokedOK)
                        throw new WSIFException(
                            "Failed to invoke method '" + fieldJavaOperationModel.getMethodName() + "'");
                } else {
                    for (int a = 0; a < fieldMethods.length; a++) {
                        if (usedOutputParam) {
                            for (int i = 0; i < fieldInParameterNames.length; i++) {
                                String outParameterName = null;
                                for (int j = 0; j < fieldOutParameterNames.length; j++) {
                                    outParameterName = fieldOutParameterNames[j];
                                    if ((outParameterName != null)
                                        && (outParameterName.equals(fieldInParameterNames[i]))) {
                                        arguments[i] = (fieldMethods[a].getParameterTypes()[i]).newInstance();
                                    }
                                }
                            }
                        }
                        try {
                            // Get a set of arguments which are compatible with the method
                            Object[] compatibleArguments =
                                getCompatibleArguments(fieldMethods[a].getParameterTypes(), arguments);
                            // If we didn't get any arguments then the parts aren't compatible with the method
                            if (compatibleArguments == null)
                                break;
                            // Parts are compatible so invoke the method with the compatible set

                            Object objRef = fieldPort.getObjectReference();
                            Trc.event(
                                this,
                                "Invoking object ",
                                objRef,
                                " method ",
                                fieldMethods[a],
                                " with arguments ",
                                compatibleArguments);

                            result =
                                fieldMethods[a].invoke(
                                    objRef,
                                    compatibleArguments);

                            Trc.event(
                                this,
                                "Returned from method, result is ",
                                result);

                            chosenMethod = fieldMethods[a];
                            invokedOK = true;
                            break;
                        } catch (IllegalArgumentException ia) {
                          	Trc.ignoredException(ia);
                            // Ignore and try next method
                        }
                    }
                    if (!invokedOK)
                        throw new WSIFException(
                            "Failed to invoke method '" + fieldJavaOperationModel.getMethodName() + "'");
                }

                String outParameterName = null;
                if (fieldOutParameterNames != null && fieldOutParameterNames.length > 0) {
                    output.setName(getOutputMessageName());
                    outParameterName = (String) fieldOutParameterNames[0];
                    if (outParameterName != null) {
                        output.setObjectPart(
                            outParameterName,
                            getCompatibleReturn(chosenMethod, result));
                        // Should we use the class of the method signature instead here ?
                    }

                    if (arguments != null) {
                        for (int i = 1; i < fieldOutParameterNames.length; i++) {
                            outParameterName = fieldOutParameterNames[i];
                            if (outParameterName != null) {
                            	try {
                            		for (int r = 0; r < fieldInParameterNames.length; r++) {
                            			if (outParameterName.equals(fieldInParameterNames[r])) {
                            				output.setObjectPart(outParameterName, arguments[r]);
                            				break;
                            			}
                            		}
                            	} catch (WSIFException e) {
                                	Trc.ignoredException(e);
                            		//ignore
                            	}
                            }
                        }
                    }
                }
            }
        } catch (InvocationTargetException ex) {
        	Trc.exception(ex);
            Throwable invocationFault = ex.getTargetException();
            String className = invocationFault.getClass().getName();
            Map faultMessageInfos = getFaultMessageInfos();
            FaultMessageInfo faultMessageInfo =
                (FaultMessageInfo) faultMessageInfos.get(className);
            if ((faultMessageInfo != null)
                && (faultMessageInfo.fieldPartName != null)) { // Found fault
                Object faultPart = invocationFault;
                // Should we use the class of the method signature here ?
                fault.setObjectPart(faultMessageInfo.fieldPartName, faultPart);
                fault.setName(faultMessageInfo.fieldMessageName);
                if (faultMessageInfo.fieldMessageName != null) {
                	Fault wsdlFault = fieldBindingOperationModel.getOperation().getFault(faultMessageInfo.fieldMessageName);
                	if (wsdlFault != null) {
                	    fault.setMessageDefinition(wsdlFault.getMessage());
                	}
                }
                operationSucceeded = false;
            } else {
                // Try to find a matching class:
                Class invocationFaultClass = invocationFault.getClass();
                Class tempClass = null;
                Iterator it = faultMessageInfos.values().iterator();
                boolean found = false;
                while (it.hasNext()) {
                    faultMessageInfo = (FaultMessageInfo) it.next();
                    try {
                        tempClass =
                            Class.forName(
                                faultMessageInfo.fieldFormatType,
                                true,
                                Thread.currentThread().getContextClassLoader());
                        if (tempClass.isAssignableFrom(invocationFaultClass)) {
                            found = true;
                            Object faultPart = invocationFault;
                            // Should we use the class of the method signature here ?
                            fault.setObjectPart(faultMessageInfo.fieldPartName, faultPart);
                            fault.setName(faultMessageInfo.fieldMessageName);
                            if (faultMessageInfo.fieldMessageName != null) {
                            	Fault wsdlFault = fieldBindingOperationModel.getOperation().getFault(faultMessageInfo.fieldMessageName);
                	            if (wsdlFault != null) {
                	            	fault.setMessageDefinition(wsdlFault.getMessage());
                	            }
                            }
                            operationSucceeded = false;
                        }
                    } catch (Exception exc) { // Nothing to do - just try the next one...
                    	Trc.ignoredException(exc);
                    }
                }
                if (!found) {
                    throw new WSIFException("Operation failed!", invocationFault);
                }
            }
        } catch (Exception ex) {
            Trc.exception(ex);

            // Log message
            MessageLogger.log(
                "WSIF.0005E",
                "Java",
                fieldJavaOperationModel.getMethodName());

            throw new WSIFException(
                this
                    + " : Could not invoke '"
                    + fieldJavaOperationModel.getMethodName()
                    + "'",
                ex);
        }

        Trc.exit(operationSucceeded);
        return operationSucceeded;
    }

    public void executeInputOnlyOperation(WSIFMessage input) throws WSIFException {

        Trc.entry(this, input);
        close();

        try {
            Object result = null;
            // Need to get the stuff here because this also initializes fieldInParameterNames
            if (fieldIsConstructor) {
                if (fieldConstructors.length <= 0)
                    throw new WSIFException("No constructors found that match the parts specified");
            } else {
                if (fieldMethods.length <= 0)
                    throw new WSIFException("No methods named '"
                        + fieldJavaOperationModel.getMethodName()
                        + "' found that match the parts specified");
            }

            Object[] arguments = null;
            Object part = null;
            if ((fieldInParameterNames != null) && (fieldInParameterNames.length > 0)) {
                arguments = new Object[fieldInParameterNames.length];
                for (int i = 0; i < fieldInParameterNames.length; i++) {
                   try {
                        part = input.getObjectPart(fieldInParameterNames[i]);
                        arguments[i] = part;
                    } catch (WSIFException e) {
                     	Trc.exception(e);
                        arguments[i] = null;
                    }
                }
            }

            boolean invokedOK = false;

            if (fieldIsConstructor) {
                for (int a = 0; a < fieldConstructors.length; a++) {
                    try {
                        // Get a set of arguments which are compatible with the ctor
                        Object[] compatibleArguments =
                            getCompatibleArguments(fieldConstructors[a].getParameterTypes(), arguments);
                        // If we didn't get any arguments then the parts aren't compatible with the ctor
                        if (compatibleArguments == null)
                            break;
                        // Parts are compatible so invoke the ctor with the compatible set

                        Trc.event(
                            this,
                            "Invoking constructor ",
                            fieldConstructors[a],
                            " with arguments ",
                            compatibleArguments);

                        result =
                            fieldConstructors[a].newInstance(
                                compatibleArguments);

                        Trc.event(
                            this,
                            "Returned from constructor, result is ",
                            result);

                        fieldPort.setObjectReference(result);
                        invokedOK = true;
                        break;
                    } catch (IllegalArgumentException ia) {
                    	Trc.ignoredException(ia);
                        // Ignore and try next constructor
                    }
                }
                if (!invokedOK)
                    throw new WSIFException("Failed to call constructor for object in Java operation");
            } else {
                if (fieldIsStatic) {
                    for (int a = 0; a < fieldMethods.length; a++) {
                        try {
                            // Get a set of arguments which are compatible with the method
                            Object[] compatibleArguments =
                                getCompatibleArguments(fieldMethods[a].getParameterTypes(), arguments);
                            // If we didn't get any arguments then the parts aren't compatible with the method
                            if (compatibleArguments == null)
                                break;
                            // Parts are compatible so invoke the method with the compatible set

                            Trc.event(
                                this,
                                "Invoking method ",
                                fieldMethods[a],
                                " with arguments ",
                                compatibleArguments);

                            result =
                                fieldMethods[a].invoke(
                                    null,
                                    compatibleArguments);

                            Trc.event(
                                this,
                                "Returned from method, result is ",
                                result);

                            invokedOK = true;
                            break;
                        } catch (IllegalArgumentException ia) {
                        	Trc.ignoredException(ia);
                            // Ignore and try next method
                        }
                    }
                    if (!invokedOK)
                        throw new WSIFException(
                            "Failed to invoke method '" + fieldJavaOperationModel.getMethodName() + "'");
                } else {
                    for (int a = 0; a < fieldMethods.length; a++) {
                        try {
                            // Get a set of arguments which are compatible with the method
                            Object[] compatibleArguments =
                                getCompatibleArguments(fieldMethods[a].getParameterTypes(), arguments);
                            // If we didn't get any arguments then the parts aren't compatible with the method
                            if (compatibleArguments == null)
                                break;
                            // Parts are compatible so invoke the method with the compatible set

                            Object objRef = fieldPort.getObjectReference();
                            Trc.event(
                                this,
                                "Invoking object ",
                                objRef,
                                " method ",
                                fieldMethods[a],
                                " with arguments ",
                                compatibleArguments);

                            result =
                                fieldMethods[a].invoke(
                                    objRef,
                                    compatibleArguments);

                            Trc.event(
                                this,
                                "Returned from method, result is ",
                                result);

                            invokedOK = true;
                            break;
                        } catch (IllegalArgumentException ia) {
                        	Trc.ignoredException(ia);
                            // Ignore and try next method
                        }
                    }
                    if (!invokedOK)
                        throw new WSIFException(
                            "Failed to invoke method '" + fieldJavaOperationModel.getMethodName() + "'");
                }
            }
        } catch (InvocationTargetException ex) {
            Trc.exception(ex);

            // Log message
            MessageLogger.log(
                "WSIF.0005E",
                "Java",
                fieldJavaOperationModel.getMethodName());

            throw new WSIFException(
                this
                    + " : Invocation of '"
                    + fieldJavaOperationModel.getMethodName()
                    + "' failed.",
                ex);
        } catch (Exception ex) {
            Trc.exception(ex);

            // Log message
            MessageLogger.log(
                "WSIF.0005E",
                "Java",
                fieldJavaOperationModel.getMethodName());

            throw new WSIFException(
                this
                    + " : Could not invoke '"
                    + fieldJavaOperationModel.getMethodName()
                    + "'",
                ex);
        }

        Trc.exit();
    }

    public String deep() {
        String buff = "";
        try {
            buff = new String(super.toString() + ":\n");
            buff += "portModel:" + Trc.brief(fieldPortModel);
            buff += " wsifPort_Java:" + fieldPort;
            buff += " bindingOperationModel:" + Trc.brief(fieldBindingOperationModel);
            buff += " JavaOperation:" + fieldJavaOperationModel;

            buff += Trc.brief("inParameterNames", fieldInParameterNames);
            buff += Trc.brief("outParameterNames", fieldOutParameterNames);          

            if (fieldFaultMessageInfos == null) {
                buff += " faultMessageInfos:null";
            } else {
                Iterator it = fieldFaultMessageInfos.keySet().iterator();
                int i = 0;
                while (it.hasNext()) {
                    String key = (String) it.next();
                    buff += " faultMessageInfos["
                        + i
                        + "]:"
                        + key
                        + " "
                        + fieldFaultMessageInfos.get(key);
                    i++;
                }
            }

            buff += Trc.brief("methods", fieldMethods);          
            buff += Trc.brief("constructors", fieldConstructors);          

            buff += " outputMessageName:" + fieldOutputMessageName;
            buff += " isStatic:" + fieldIsStatic;
            buff += " isConstructor:" + fieldIsConstructor;

            if (fieldTypeMaps == null) {
                buff += " faultTypeMaps:null";
            } else {
                Iterator it = fieldTypeMaps.keySet().iterator();
                int i = 0;
                while (it.hasNext()) {
                    QName key = (QName) it.next();
                    buff += " typeMaps[" + i + "]:" + key + " " + fieldTypeMaps.get(key);
                    i++;
                }
            }
        } catch (Exception e) {
            Trc.exceptionInTrace(e);
        }

        return buff;
    }
}