/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:  
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "WSIF" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 2001, 2002, International
 * Business Machines, Inc., http://www.apache.org.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

package org.apache.wsif.util.jms;

import java.util.ArrayList;
import java.util.Arrays;

import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.Queue;
import javax.jms.QueueReceiver;
import org.apache.wsif.WSIFException;
import org.apache.wsif.logging.Trc;

/**
 * Server destination for the JMS2HTTPBridge
 * @author Mark Whitlock <whitlock@apache.org>
 */
public class JMS2HTTPBridgeDestination extends WSIFJMSDestination {
    public static final String COLDSTART = "cold";
    public static final String WARMSTART = "warm";

    private boolean verbose;
    private static final ArrayList allStarts =
        new ArrayList(Arrays.asList(new Object[] { COLDSTART, WARMSTART }));

    /**
     * Public constructor.
     * @param finder used to find JMS objects.
     * @param altdestName is an alterative JMS provider destination name
     * @param timeout is the maximum time to wait on a synchronous receive
     * @param startType is WARMSTART or COLDSTART. Cold means wipe the read queue on startup.
     */
    public JMS2HTTPBridgeDestination(
        WSIFJMSFinder finder,
        String altDestName,
        long timeout,
        String startType,
        boolean verbose)
        throws WSIFException {
        super(finder, altDestName, timeout);
        Trc.entry(
            this,
            finder,
            altDestName,
            new Long(timeout),
            startType,
            new Boolean(verbose));

        this.verbose = verbose;
        
        // Swap the queues because we're a server not a client.  	
        readQ = writeQ;
        writeQ = null;

        if (!allStarts.contains(startType))
            throw new WSIFException("StartType must either be warm or cold");

        if (COLDSTART.equals(startType)) {
            if (verbose)
                System.out.println("Wiping messages off the read queue");
            Message msg = null;
            try {
                QueueReceiver rec = session.createReceiver(readQ);
                for (;;) {
                    msg = rec.receive(100);
                    if (msg != null) {
                        if (verbose)
                            System.out.println("Removing an input message");
                    } else
                        break;
                }
            } catch (Exception ignored) {
            	Trc.exception(ignored);
            }
        }
        Trc.exit();
    }

    /**
     * Create a listener thread to listen for messages. This waits forever until it gets 
     * an InterruptedException. This listens on the read queue.
     * @param listener is the JMS message and exception callback interface implementation
     */
    public void listen(WSIFJMSListener listener) throws WSIFException {
        Trc.entry(this, listener);
        listen(listener, readQ);
        Trc.exit();
    }

    /**
     * Create a listener thread to listen for messages. This waits forever until it gets 
     * an InterruptedException. 
     * @param listener is the JMS message and exception callback interface implementation
     * @param queue to listen on
     */
    public void listen(WSIFJMSListener listener, Queue queue)
        throws WSIFException {
        Trc.entry(this, listener, queue);
        areWeClosed();

        try {
            QueueReceiver qr = session.createReceiver(queue);
            qr.setMessageListener(listener);
            connection.setExceptionListener(listener);

            connection.start();

            for (int i = 1; !Thread.interrupted(); i++) {
                Thread.yield();
                Thread.sleep(5000);
                if (verbose)
                    System.out.println("Waiting... " + i);
            }
        } catch (JMSException je) {
        	Trc.exception(je);
            throw WSIFJMSConstants.ToWsifException(je);
        } catch (InterruptedException ignored) {
        	Trc.exception(ignored);
            System.out.println("Exitting");
        }
        Trc.exit();
    }

    // Stop anyone overwriting our readQ
    public void setReplyToQueue() throws WSIFException {
        Trc.entry(this);
        Trc.exit();
    }

    public void setReplyToQueue(String replyTo) throws WSIFException {
        Trc.entry(this, replyTo);
        Trc.exit();
    }

    /**
     * Set the replyTo queue. Special bridge version.
     * @param replyTo queue.
     */
    public void setReplyToQueue(Queue replyTo) throws WSIFException {
        Trc.entry(this, replyTo);
        areWeClosed();

        if (writeQ == null)
            writeQ = replyTo;
        else if (!writeQ.equals(replyTo)) {

            if (sender != null) {
                try {
                    sender.close();
                } catch (Exception e) {
                	Trc.exception(e);
                }
                sender = null;
            }
            writeQ = replyTo;
        }
        Trc.exit();
    }
}