/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:  
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "WSIF" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 2001, 2002, International
 * Business Machines, Inc., http://www.apache.org.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

package util;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Stack;
import java.util.Vector;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;
import junit.textui.TestRunner;

/** 
 * A JUnit test to check our files for the dreaded ^M character. The test displays a list of
 * all the WSIF files it finds that contain the character.
 * 
 * This test is configured by the "fix" component setting in the wsif.test.properties file.
 * If fix is set to on, this test will attempt to remove any ^M chars it finds in the WSIF files.
 * It is recommended that fix=on if only set if you are sure that the ^M characters are present
 * in the server copy of the file.
 *  
 * @author Owen Burroughs <owenb@apache.org>
 */
public class FileCharacterTest extends TestCase {

    static Stack dirStack = new Stack();
    static Vector files = new Vector();
    static Vector badFiles = new Vector();

    public FileCharacterTest(String name) {
        super(name);
    }

    public static void main(String[] args) {
        TestRunner.run(suite());
    }

    public static Test suite() {
        return new TestSuite(FileCharacterTest.class);
    }

    public void testFiles() {
        File path = null;
        int index;
        try {
            String temp = TestUtilities.getWsdlPath("");
            path = new File(temp);

            //check that path exists 
            if (!path.exists())
                throw new FileNotFoundException();

            //if path is not a directory, check that it is a file 
            if (!path.isDirectory())
                //if path is a file, get parent directory (ignore file name) 
                if (path.isFile())
                    path = path.getParentFile();
            //else report error and exit 
            else {
                System.out.println("\npath '" + path + "' is not a file or directory!");
                System.exit(1);
            }

            //if path is unreadable, report error and exit 
            if (!path.canRead()) {
                System.out.println("FCT: " + path + ": Permission denied");
                System.exit(1);
            }
        }
        //catch non-existent file error 
        catch (FileNotFoundException e) {
            System.out.println("\npath '" + path + "' does not exist!");
            System.exit(1);
        }

        findFiles(path);
        while (!dirStack.empty()) {
            dirStack.pop();
        }
        int total = files.size();
        Enumeration e = files.elements();
        while (e.hasMoreElements()) {
            File f = (File) e.nextElement();
            //System.out.println(f); 
            checkFile(f);
        }

        int bad = badFiles.size();
        if (bad > 0) {
            System.out.println(
                "\nOut of the "
                    + total
                    + " WSIF files under "
                    + path
                    + ",\nthe following "
                    + bad
                    + " files contain ^M characters:\n");
            Enumeration e2 = badFiles.elements();
            while (e2.hasMoreElements()) {
                File f = (File) e2.nextElement();
                System.out.println(f);
            }

            if (TestUtilities.areWeTesting("fix")) {
	            System.out.println("\nNow to fix them :-)\n");
	            Enumeration e3 = badFiles.elements();
	            while (e3.hasMoreElements()) {
	                File f = (File) e3.nextElement();
	                fixFile(f);
	            }
            }
        } else {
            System.out.println("All files under " + path + " are OK :-)");
        }
    }

    private void findFiles(File path) {
        try {
            //obtain canonical file (condenses cycles e.g. from links) 
            File canonPath = path.getCanonicalFile();
            FileCharacterTest.dirStack.push(canonPath.toString());
        } catch (IOException e) {
            //trivially handle exception and carry on 
            System.out.println("Canonical problem");
        }
        //search directory for file names that satisfy FilenameFilter 
        //select a subset of the directory's filenames 
        File[] temp = path.listFiles(new FilenameFilter() {
        
            public boolean accept(File afile, String name) {
                //create a new file object for each sub-object 
                File newFile = new File(afile, name);

                //ignore if file object is unreadable 
                if (!newFile.canRead()) {
                    //report omission 
                    System.out.println(newFile + ": Permission denied");
                    return false;
                }

                //ignore files in a bin directory. We are only interested in source 
                if (newFile.getPath().indexOf("\\bin\\") != -1) {
                    return false;
                }

                //ignore class files!!!! 
                if (name.toUpperCase().endsWith(".CLASS")) {
                    return false;
                }

                //ignore ear files!!!! 
                if (name.toUpperCase().endsWith(".EAR")) {
                    return false;
                }

                //ignore jar files!!!! 
                if (name.toUpperCase().endsWith(".JAR")) {
                    return false;
                }

                //ignore zip files!!!! 
                if (name.toUpperCase().endsWith(".ZIP")) {
                    return false;
                }

                //ignore war files!!!! 
                if (name.toUpperCase().endsWith(".WAR")) {
                    return false;
                }

                //ignore eclipse files!!!! 
                if (name.toUpperCase().startsWith(".VCM")
                    || name.toUpperCase().startsWith(".CLASSPATH")) {
                    return false;
                }

                //ignore if file object is hidden 
                if (newFile.isHidden())
                    return false;
                return !(dirStack.contains(newFile.toString()));
            }
        });
        for (int a = 0; a < temp.length; a++) {
            if (temp[a].isDirectory()) {
                findFiles(temp[a]);
                //sub-tree traversed - pop object from stack 
                dirStack.pop();
            } else {
                files.add(temp[a]);
            }
        }
    }

    private void checkFile(File file) {
        FileReader fr = null;
        try {
            fr = new FileReader(file);
            char[] chunk = new char[500];
            int ok = fr.read(chunk);
            while (ok != -1) {
                String s = new String(chunk);
                if (s.indexOf("\r") != -1) {
                    badFiles.add(file);
                    break;
                }
                ok = fr.read(chunk);
            }
            fr.close();
        } catch (Exception e) {
            System.out.println("Unable to check file " + file + " : " + e);
            assertTrue(false);
        }
    }

    private void fixFile(File file) {
        FileReader fr = null;
        StringBuffer sb = new StringBuffer((int) file.length());
        try {
            fr = new FileReader(file);
            char[] chunk = new char[(int) file.length()];
            int ok = fr.read(chunk);
            while (ok != -1) {
                String s = new String(chunk);
                //s = s.replace('\r', ' '); 
                sb.append(s);
                ok = fr.read(chunk);
            }
            fr.close();

            // Not amazingly efficient but the for the best end result
            // iterate through all chars and when we find a ^M, delete it.
            // The alternative replace methods leave us with a ' ' char
            // at the end of lines which mucks up properties files!!            
            for (int x = 0; x < sb.length(); x++) {
                if (sb.charAt(x) == '\r') {
                    sb.deleteCharAt(x);
                    if (x > 0)
                        x--;
                }
            }

        } catch (Exception e) {
            System.out.println("Unable to read file " + file + " : " + e);
            assertTrue(false);
        }
        try {
            FileWriter fw = new FileWriter(file);
            fw.write(sb.toString().trim());
            fw.flush();
            fw.close();
            System.out.println("Fixed file " + file);
        } catch (Exception e) {
            System.out.println("Unable to write file " + file + " : " + e);
            assertTrue(false);
        }
    }
}
